/************************************************************************

   vobcut v1.0, (c) by Bernhard Oemer <oemer@tph.tuwien.ac.at>, 2009
   
   analyse and cut mpeg2 program streams without decoding and 
   demultiplexing
    
   features:
   
   - locates the starts of mpeg2 sequences and their
     associated audio packages in an mpeg2 program stream.

   - command line interface, run vobcut -h for details.

   - identifies and makes it easy to extract and concatenate
     homgenous chunks which share the same format.

   - hard format features include:

     * mpeg video format (esp. switches between 4:3 and 16:9)
       by decoding mpeg SEQ, GOP and 
     * mpeg audio format (usually remains constant for a channel)

   - soft format features include:

     * letterbox (i.e. black bars to widen the aspect ratio)
       by comparing the encoded size of the outer scanlines to
       the inner scanlines. except for finding the slice header,
       no decoding takes place.
     * stereo encoded mono (mp2 audio only) by partially
       decoding the mp2 stream and comparing the spectra of
       left and right channel.

     reliability depends on content and also varies considerably 
     between channels, so the heuristics can require tuning or
     might even fail. Generally, when in doubt (e.g. dark scenes, 
     silence) the sequence gets classified als lbox or stmono.

   - rewriting of CSR, PTS, DTS and ESCR timestamps to avoid
     discontinuities and clicks when cutting out scenes

   - trimming of incomplete audio packed to avoid clicks by
     invalidating the header prefix, so players ignore them.
     (while this violates the standard, it works just fine)

   - supports caching of block information to speed up subsequent 
     calls on the same file

   - small, fast and standalone (no external decoders, etc.)

   - uses only standard includes and standard c++ libraires

   installation:
   
   g++ -O2 vobcut.cc -o /usr/local/bin/vobcut -lm

   limitations:
   
   I have written the program to cut PAL DVB-S program streams from
   my sat receiver after having ruined some recordings with a
   graphical video editing program (loss of A/V-sync). While it might 
   work with other mpeg2 PSs, this is not tested.
   
   The philosophy of vobcut is to change as little as possible from the
   original stream and avoid repacking, so no errors are repaired and
   concatenations lead to incomplete audio packets as - other than the
   start of sequences - audio packets are not pack-aligned.
   
   vobcut only deals with the first mpeg2 video (0xe0) and audio (0xc0)
   stream and expects those to be present.
   
   letterbox detection assumes that slices are top down, 16-pixel wide 
   and cover the full width, which this is not required by the standard. 
   It also assumes that the black bands compress significantly better 
   than the rest, which is not always the case.
   
   audio has to mpeg audio, non-mpeg audio in private streams is not
   supported. stereo encoded mono detection only works with mp2.
   
   lbox and stmono detection produce lots of false positives so they are 
   really only useful if you want to extract (as opposed to cutting away)
   letterboxed or mono content.
   
   the default scanning strategy which consists of many short scans with
   long gaps in between which only get read when necessary might not only
   overlook scenes but can in fact take longer than a complete scan (option
   -e). This happens esp. when doing lbox/stmono detection is active on
   content that is mostly full-screen/stereo.

   licence:
  
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   
   See http://www.gnu.org/licenses/gpl-3.0.txt for mor details.

************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <string>
#include <vector>
#include <set>
#include <map>

using namespace std;

typedef u_int8_t u8;
typedef u_int16_t u16;
typedef u_int32_t u32;
typedef u_int64_t u64;
typedef int64_t i64;

int verbose=1;
int quiet=0;
int calc_timestamps=1;
int trim_audio=1;
int audio_only=0;
long pack_size=0x800;
int dump=0;
double iscan_len=3;
double iscan_ival=30;
long iscans=0;
const int min_iscans=10;
double vote_window=5;
double preview_len=5;
int preview=0;
int view_bound=0;
i64 do_trunc=-1;

int chk_lbox=-1;
int chk_mono=-1;
int lbox_thres=50;
const int lbox_slices=2;
int sample_mono_thres=1;
double mono_thres=60;
double mono_limit;
double remove_limit=10;

double adj_a=0;
double adj_b=0;

const char *ofile=0;
int read_cachefile=1;
int write_cachefile=1;
int create_cachedir=0;

int write_cache=1;
set<int> chunks;
vector<long> cuts;
vector<double> tcuts;
int plot=0;

#define CACHE_DIR ".vobcut"
#define CACHE_PREFIX CACHE_DIR "/"
#define CACHE_POSTFIX ".cache"
#define CACHE_MAGIC 0x0a0d747543626f56ull

void usage() {
  fprintf(stderr,
    "USAGE: vobcut [options] <mpeg-file> ... (normal mode)\n"
    "       vobcut -D <mpeg-file> ...        (dump mode)\n"
    "       vobcut -T <blocks> <mpeg-file>   (truncate file)\n"
    "analyse and cut mpeg2 program streams without decoding and demultiplexing\n"
    "version 1.0, (c) by Bernhard Oemer, 2009, licence: GPLv3\n\n"
    "general options:\n"
    "-h         help (this message)\n"
    "-v         verbose messages\n"
    "-Q         supress warnings\n"
    "-t         disable timestamp recalculation\n"
    "-k         keep incomplete mp2 audio packages\n"
    "-P <len>   mpeg PS pack size in bytes (%ld)\n"
    "-D         dump sequences and audio packets and exit\n"
    "-T <n>     truncate single file to n blocks (dangerous) and exit\n"
    "-f <time>  force cut around PTS <time> given in seconds\n"
    "-F <block> force cut at next sequence after the given block\n"
    "I/O, scanning and caching options:\n"
    "-c / -C    do not read / write cache file even if .vobcut dir exists\n"
    "-d         create .vobcut cache directory if nonexistent\n"
    "-o <file>  output file, use -o- to write to stdout\n"
    "-0 ... -9  write chunk number 0...9 (default: largest)\n"
    "-p [len]   preview: write first and last [len] seconds to stdout (%g)\n"
    "-q [len]   preview boundaries enclosing selected chunks\n"
    "-a <offs>  soft adjust start (no hard transitions) *\n"
    "-b <offs>  soft adjust end (no hard transitions) *\n"
    "-s <secs>  min. initial partial scan sample length in seconds (%g)\n"
    "-i <secs>  max. initial partial scan sample interval in seconds (%g)\n"
    "-n <n>     number of initial scans, use -v to check block ranges (auto)\n"
    "-e         always scan entire file; can be faster than default partial scans\n"
    "soft format (lbox/stmono) detection options:\n"
    "-l / -L    disable/enable letterbox recognition (auto)\n"
    "-m / -M    disable/enable stereo encoded mono recognition (auto)\n"
    "-w <secs>  letterbox/mono vote window in seconds (%g)\n"
    "-r <secs>  removal limit for insular soft sections in seconds (%g) *\n" 
    "-B <perc.> threshold for letterbox slice size in percent of inner slices (%d)\n"
    "-S <thres> threshold for stereo encoded mono (auto, fallback value %g)\n"
    "* a hard section/transition implies an mpeg format change (e.g. 4:3 -> 16:9)\n"
    ,pack_size,preview_len,iscan_len,iscan_ival,vote_window,remove_limit,lbox_thres,mono_thres
  );
  exit(1);
}

FILE *dmp_file=stdout;

void dmp(int level,const char *fmt,...) {
  va_list ap;
  if(!quiet && level<=verbose) {
    va_start(ap, fmt);
    vfprintf(dmp_file,fmt,ap);
    va_end(ap);
  }
}

void warn(const char *fmt,...) {
  va_list ap;
  fflush(0);
  if(!quiet) {
    va_start(ap, fmt);
    vfprintf(stderr,fmt,ap);
    va_end(ap);
  }
  fflush(0);
}

void parse_options(int argc,char **argv) {
  int c;
  char *pc;
  double val;

  while(1) {
    c=getopt(argc,argv,"hvQAP:Di:s:ew:B:S:lLmMo:p::q::a:b:n:0123456789T:r:cCdtkf:F:x");
    if(c==-1) break;
    errno=0;
    switch(c) {
      case 'o':
        ofile=optarg;
        break;
      case 'P':
        pack_size=strtol(optarg,&pc,0);
        if(*pc || errno) error(1,0,"invalid packsize -%c %s.\n",c,optarg);
        break;
      case 'B':
        lbox_thres=strtol(optarg,&pc,0);
        if(*pc || errno) error(1,0,"invalid letterbox slice size percentage -%c %s.\n",c,optarg);
        break;
      case 'S':
        mono_thres=strtol(optarg,&pc,0);
        sample_mono_thres=0;
        if(*pc || errno) error(1,0,"invalid threshold -%c %s.\n",c,optarg);
        break;
      case 's':
        iscan_len=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_len<0) error(1,0,"invalid sample length -%c %s.\n",c,optarg);
        break;
      case 'e':
        iscan_len=0;
        break;
      case 'i':
        iscan_ival=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_ival<0) error(1,0,"invalid sample interval -%c %s.\n",c,optarg);
        break;
      case 'n':
        iscans=strtol(optarg,&pc,0);
        if(*pc || errno || iscans<1) error(1,0,"invalid number of scans -%c %s.\n",c,optarg);
        break;
      case 'w':
        vote_window=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_ival<0) error(1,0,"invalid vote window -%c %s.\n",c,optarg);
        break;
      case 'r':
        remove_limit=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_ival<0) error(1,0,"invalid removal limit -%c %s.\n",c,optarg);
        break;
      case 'a':
        adj_a=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_ival<0) error(1,0,"invalid adjustment time -%c %s.\n",c,optarg);
        break;
      case 'b':
        adj_b=(double)strtod(optarg,&pc);
        if(*pc || errno || iscan_ival<0) error(1,0,"invalid adjustment time -%c %s.\n",c,optarg);
        break;
      case 'q':
        view_bound=1;
      case 'p':
        preview=1;
        dmp_file=stderr;
        if(optarg) {
          preview_len=(double)strtod(optarg,&pc);
          if(*pc || errno || iscan_ival<0) error(1,0,"invalid preview length -%c %s.\n",c,optarg);
        }
        break;
      case 'v':
        verbose++;
        break;
      case 'Q':
        quiet++;
        verbose=0;
        break;
      case 't':
        calc_timestamps=0;
        break;
      case 'k':
        trim_audio=0;
        break;
      case 'D':
        dump=1;
        break;
      case 'l':
        chk_lbox=0;
        break;
      case 'm':
        chk_mono=0;
        break;
      case 'L':
        chk_lbox=1;
        break;
      case 'M':
        chk_mono=1;
        break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        chunks.insert(c-'0');
        break;
      case 'T':
        do_trunc=strtol(optarg,&pc,0);
        if(*pc || errno || do_trunc<0) error(1,0,"invalid truncation length -%c %s.\n",c,optarg);
        break;
      case 'c':
        read_cachefile=0;
        break;
      case 'C':
        write_cachefile=0;
        break;
      case 'd':
        create_cachedir=1;
        break;
      case 'F':
        {
          long b=strtol(optarg,&pc,0);
          if(*pc || errno || b<=0) error(1,0,"invalid cut block -%c %s.\n",c,optarg);
          cuts.push_back(b);
        }
        break;
      case 'f':
        {
          double t=(double)strtod(optarg,&pc);
          if(*pc || errno || t<=0) error(1,0,"invalid cut time -%c %s.\n",c,optarg);
          tcuts.push_back(t);
        }
        break;
      case 'x':
        dmp_file=stderr;
        plot=1;
        break;
      case 'h':
      default:
        usage();
    }
  }
  mono_limit=0.00001*mono_thres*mono_thres;
}

const int   audio_br[4][16] = {
        { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
        { 0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,-1 },
	{ 0,32,48,56,64,80,96,112,128,160,192,224,256,320,384,-1},
        { 0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,-1}
};
const int audio_freq[4] = { 44100,48000,32000,0 };
const int vframe_dur[16] = { 0,3754,3750,3600,3003,3000,1800,1501,1500,0,0,0,0,0,0,0 };
const int aframe_dur[4] = { 2351,2160,3240,0 };

class vob_container {
  struct vob_t {
    const char *fname;
    int fd;
    struct stat st;
    long blocks;
    long boffs;
  };
  map<long,vob_t> vobs;
  vector<u64> vmask;
  u8 *buffer;
public:
  string cachefile() const {
    if(vobs.empty()) return string("");
    map<long,vob_t>::const_iterator it=vobs.begin();
    string s=CACHE_PREFIX;
    s+=it->second.fname;
    for(it++;it!=vobs.end();it++) {
      s+="+";
      s+=it->second.fname;
    }
    s+=CACHE_POSTFIX;
    return s;
  };
  string filename() const {
    if(vobs.empty()) return string("");
    map<long,vob_t>::const_iterator it=vobs.begin();
    string s=it->second.fname;
    for(it++;it!=vobs.end();it++) {
      s+="+";
      s+=s+=it->second.fname;
    }
    return s;
  };
  long n_loaded;
  long n_visited;
  long blocks() const { return vobs.empty() ? 0 : vobs.rbegin()->first; }
  double covered() const { return n_visited/(double)blocks(); }
  void visit_block(long i) { 
    map<long,vob_t>::const_iterator it=vobs.upper_bound(i);
    if(it==vobs.end()) error(1,0,"nonexistant block %ld requested for caching\n",i);
    if(!(vmask[i>>6]&(1ull<<(i&63)))) {
      vmask[i>>6]|=(1ull<<(i&63));
      n_loaded++;
    }
    n_visited++;
  }
  u8 *block(long i) { 
    map<long,vob_t>::const_iterator it=vobs.upper_bound(i);
    if(it==vobs.end()) error(1,0,"nonexistant block %ld requested\n",i);
    if(!(vmask[i>>6]&(1ull<<(i&63)))) {
      vmask[i>>6]|=(1ull<<(i&63));
      n_loaded++;
    }
    n_visited++;
    const vob_t& v=it->second;
    off_t o=(i-v.boffs)*pack_size;
    if(lseek(v.fd,o,SEEK_SET)!=o) 
      error(1,errno,"cannot seek to %ld in %s\n",o,v.fname);
    if(read(v.fd,buffer,pack_size)!=pack_size) 
      error(1,errno,"read error at %ld in %s\n",o,v.fname);
    return buffer; 
  }
  u64 visited(long i) const { return vmask[i>>6] & (1ull<<(i&63)); }
  const u8 *data(long i) {
    const u8 *p=block(i);
    if(*(u32*)p!=0xba010000) error(1,0,"invalid pack header at block %ld\n",i);
    int stuff=p[13]&7;
    return p+14+stuff;
  }
  void add(const char *fname) {
    if(!buffer) {
      buffer=new u8[pack_size];
      if(!buffer) error(1,0,"cannot allocate buffer\n");
    }
    vob_t v;
    if((v.fd=open(fname,O_RDONLY))<0 || fstat(v.fd,&v.st)) error(1,errno,"cannot open %s\n",fname);
    if(!v.st.st_size || v.st.st_size%pack_size && !audio_only) error(1,0,"file %s has invalid size.\n",fname);
    v.fname=fname;
    v.blocks=v.st.st_size/pack_size;
    v.boffs=blocks();
    vobs[v.boffs+v.blocks]=v;
    while(vmask.size()*64<v.boffs+v.blocks) vmask.push_back(0);
  }
  vob_container() : buffer(0),n_loaded(0),n_visited(0) { }
  ~vob_container() {
    for(map<long,vob_t>::iterator it=vobs.begin();it!=vobs.end();it++) {
      if(close(it->second.fd)) error(1,errno,"cannot close %s\n",it->second.fname);
    }
    if(buffer) delete[] buffer;
  }
};

#define MP_STEREO 0
#define MP_JSTEREO 1
#define MP_DUAL 2
#define MP_MONO 3

struct blockinfo_t {
  u8 sid;	// video: 0xE0-0xEF			audio: 0xC0-0xDF, (0xBD AC3)
  u8 cnt;	// video: numb. of slices read   	audio: number of audio blocks
  u16 val;	// video: lbox value			audio: stereo value
  u32 form;	// video: SEQ header bytes 4-7		audio: 16 bit MPEG header
  u32 scr;	// standard clock rate
  u32 pts;	// presentation time stamp for PIC #0 in SEQ or 1st audio frame
  u32 mux;	// MUX rate in bytes/s
  int is_audio() const { return (sid&0xe0)==0xc0 || sid==0xbd; }
  int is_video() const { return (sid&0xe0)==0xe0; }
  int is_av() const { return is_video() || is_audio(); }
  int is_seq() const { return is_video() && form; }
  int is_as() const { return is_audio() || is_seq(); }
  int is_lbox() const { return chk_lbox && is_video() && cnt && val<lbox_thres; }
  int is_stmono() const { return chk_mono && is_audio() && (form&0xc0)!=0xc && val<mono_thres; }
  int is_special() const { return is_lbox() || is_stmono(); }
  int scanlines() const { return is_seq() ? form>>12&0xff : 0; }
  u32 format() const { return is_audio() ? form&0xfffffccf : form; }
  blockinfo_t() : sid(0),cnt(0),val(0),form(0),scr(0),pts(0),mux(0) { };
};    

 
size_t get_audio_len(u32 hd) { // size of mpeg audio frame including header
  unsigned lay=4-((hd>>17)&0x3);
  unsigned br=(hd>>12)&0xf;
  unsigned fr=(hd>>10)&0x3;
  unsigned pad=(hd>>9)&0x1;
  long rate=audio_br[lay][br];
  long freq=audio_freq[fr];
  if(rate<=0 || freq<=0) return 0;
  switch(lay) {
    case 1:
      return ((rate*12000)/freq+pad)*4;
    case 2:
    case 3:
      return (rate*144000)/freq+pad;
    default:
      return 0;
  }
}

const char *ahd_str(u32 hd) {
  if(hd<0xfff00000) return "unknown_audio_format";
  static int ibuf=0;
  static char buf[8][64];
  unsigned lay=4-((hd>>17)&0x3);
  unsigned crc=!((hd>>16)&0x1);
  unsigned br=(hd>>12)&0xf;
  unsigned fr=(hd>>10)&0x3;
  unsigned pad=(hd>>9)&0x1;
  unsigned priv=(hd>>8)&0x1;
  unsigned mode=(hd>>6)&0x3;
  unsigned mext=(hd>>4)&0x3;
  unsigned copy=(hd>>3)&0x1;
  unsigned orig=(hd>>2)&0x1;
  unsigned emph=hd&0x3;
  int rate=audio_br[lay][br];
  int freq=audio_freq[fr];
  ibuf=(ibuf+1)&7;
  static const char *Lay[4] = { "MPX","MP1","MP2","MP3" };
  static const char *Crc[2] = { "",",crc" };
  static const char *Pad[2] = { "",",pad" };
  static const char *Priv[2] = { "",",priv" };
  static const char *Copy[2] = { "",",copy" };
  static const char *Orig[2] = { "",",orig" };
  static const char *Emph[4] = { "",",50/15","emph2","ccitt" };
  static const char *Mode[4] = { "stereo","joint","dual","mono" };
  snprintf(buf[ibuf],64,"%s %d-%d %s ext%d%s%s%s%s%s%s",
           Lay[lay],freq,rate,Mode[mode],mext,Crc[crc],
           Copy[copy],Orig[orig],Emph[emph],Priv[priv],Pad[pad]);
  return buf[ibuf];
}

const char *vhd_str(u32 hd) {
  static int ibuf=0;
  static char buf[8][64];
  unsigned x=hd>>20;
  unsigned y=hd>>8&0xfff;
  unsigned ar=hd>>4&0xf;
  unsigned fr=hd&0xf;
  ibuf=(ibuf+1)&7;
  static const char *Ar[16] = { "err","1:1","4:3","16:9","2.21:1","AR5","AR6","AR7","AR8","AR9","AR10","AR11","AR12","AR13","AR14","AR15" };
  static const char *Fr[16] = { "err","23.976","24","25","29.97","30","50","59.94","60","res9","res10","res11","res12","res13","res14","res15" };
  snprintf(buf[ibuf],64,"%dx%d %s %s fps",x,y,Ar[ar],Fr[fr]);
  return buf[ibuf];
}

struct format_t {
  static int ibuf;
  static char buf[8][128];
  u32 af;
  u32 vf;
  bool lbox;
  bool stmono;
  bool operator == (const format_t& f) const { 
    return af==f.af && vf==f.vf && lbox==f.lbox && stmono==f.stmono; 
  }
  bool operator != (const format_t& f) const { return !((*this)==f); }
  const char *str() {
    ibuf=(ibuf+1)&7;
    snprintf(buf[ibuf],128,"V: %s%s; A: %s%s",vhd_str(vf),lbox?" letterbox":"",ahd_str(af),stmono ? " stmono":"");
    return buf[ibuf];
  }
};

bool sim(const format_t& a,const format_t& b) { return a.af==b.af && a.vf==b.vf; }

int format_t::ibuf=0;
char format_t::buf[8][128];

template <int n>
inline u8 getbits(const u8 *&p,int &i) {
  u32 r=((p[0]<<8|p[1])<<i&0xffff)>>(16-n);
  i+=n; p+=i>>3; i&=7;
  return r;
}


double get_stereo_factor(u32 hd,const u8* p) { // *p to point at first byte after the 32 bit header
  unsigned lay=4-((hd>>17)&0x3);
  unsigned crc=!((hd>>16)&0x1);
  unsigned br=(hd>>12)&0xf;
  unsigned fr=(hd>>10)&0x3;
  unsigned mode=(hd>>6)&0x3;
  unsigned mext=(hd>>4)&0x3;
  int q=0;

  if(lay!=2 || fr!=1 || br<3 || mode!=MP_MONO && br<7) { // only mp2, 48 kHz and >= 56 kbit/s*nch 
    warn("no MP2 48 kHz and >= 56 kbit/s per channel\n");
    return -1;
  }
  if(crc) p+=2;
  if(mode==MP_MONO) return HUGE_VAL;
  
  const int alloc4=11;
  const int alloc3=23;
  const int sblimit=27;
  const int sblimit2=2*sblimit;
  const int freqlimit=20;
  const int freqlimit2=2*freqlimit;
  const int noiselimit=0x2f;
  const int maxdiff=3;

  static const double diff_factor[64] = { // heurisic weights, 1/2 sqrt(ampl)
    1.00000,0.94387,0.89090,0.84090,0.79370,0.74915,0.70711,0.66742,0.62996,0.59460,0.56123,0.52973,0.50000,0.47194,0.44545,0.42045,
    0.39685,0.37458,0.35355,0.33371,0.31498,0.29730,0.28062,0.26487,0.25000,0.23597,0.22272,0.21022,0.19843,0.18729,0.17678,0.16685,
    0.15749,0.14865,0.14031,0.13243,0.12500,0.11798,0.11136,0.10511,0.09921,0.09364,0.08839,0.08343,0.07875,0.07433,0.07015,0.06622,
    0.06250,0.05899,0.05568,0.05256,0.04961,0.04682,0.04419,0.04171,0.03937,0.03716,0.03508,0.03311,0.03125,0.02950,0.02784,0.02628
  };

  int jsbound = mode==MP_JSTEREO ? mext*4+4 : sblimit;
  int abit[sblimit2];
  int scfsi[sblimit2];
  int sf[sblimit2][3];
  {  
    int sb=0;
    for(;sb<alloc4;sb++)  { abit[2*sb]=abit[2*sb+1]=getbits<4>(p,q); if(sb<jsbound) abit[2*sb+1]=getbits<4>(p,q); }
    for(;sb<alloc3;sb++)  { abit[2*sb]=abit[2*sb+1]=getbits<3>(p,q); if(sb<jsbound) abit[2*sb+1]=getbits<3>(p,q); }
    for(;sb<sblimit;sb++) { abit[2*sb]=abit[2*sb+1]=getbits<2>(p,q); if(sb<jsbound) abit[2*sb+1]=getbits<2>(p,q); }
  }
  for(int i=0;i<sblimit2;i++) 
    if(abit[i]) scfsi[i]=getbits<2>(p,q);
  for(int i=0;i<freqlimit2;i++) {
    if(abit[i]) {
      switch(scfsi[i]) {
        case 0: // 1+1+1
          sf[i][0]=min<int>(getbits<6>(p,q),noiselimit);
          sf[i][1]=min<int>(getbits<6>(p,q),noiselimit);
          sf[i][2]=min<int>(getbits<6>(p,q),noiselimit);
          break;
        case 1: // 2+1
          sf[i][0]=sf[i][1]=min<int>(getbits<6>(p,q),noiselimit);
          sf[i][2]=min<int>(getbits<6>(p,q),noiselimit);
          break;
        case 2: // 3
          sf[i][0]=sf[i][1]=sf[i][2]=min<int>(getbits<6>(p,q),noiselimit);
          break;
        case 3: // 1+2
        default:
          sf[i][0]=min<int>(getbits<6>(p,q),noiselimit);
          sf[i][1]=sf[i][2]=min<int>(getbits<6>(p,q),noiselimit);
          break;
      }
    } else {
      sf[i][0]=sf[i][1]=sf[i][2]=noiselimit;
    }
  }
  int ref=0;
  for(int l=0,r=1;l<freqlimit2/3;l+=2,r+=2) ref+=sf[l][1]+sf[r][1];
  ref/=freqlimit;

  double diff=0;
  for(int l=0,r=1;l<freqlimit2;l+=2,r+=2) {
    if(abit[l] && abit[r]) {
      for(int j=0;j<3;j++) {
        int a=min<int>(sf[l][j],sf[r][j]);
        int d=min<int>(maxdiff,abs(sf[l][j]-sf[r][j]));
        diff+=d*diff_factor[max<int>(0,a-ref)];
      }
    }
  }
  return diff/diff_factor[noiselimit];
}

#define SCR_MASK  0x3bfffbfff8000000ull
#define PTS_MASK  0x0efffefffe000000ull
#define DTS_MASK  0x0efffefffe000000ull
#define ESCR_MASK 0x3bfffbfff8000000ull

i64 unpack(const u8* p,u64 mask) {
  i64 t=0;
  u8 m=0x80;
  while(mask) {
    if(mask&(1ull<<63)) {
      t<<=1;
      if(*p&m) t++;
    }
    mask<<=1;
    m>>=1;
    if(!m) { m=0x80; p++; }
  }
  return t;
}

void pack(u8* p,u64 mask,i64 t) {
  if(!mask) return;
  u8 m=0x01;
  p+=7;
  while(!(mask&0xff)) { mask>>=8; p--; } 
  while(mask) {
    if(mask&1) {
      *p&=~m;
      if(t&1) *p|=m;
      t>>=1;
    }
    mask>>=1;
    m<<=1;
    if(!m) { m=0x01; p--; }
  }
}

void shift_block(u8* p,long delta) {
  if(!delta) return;
  const u8 *pend=p+pack_size-24;
  for(;p<pend;p++) {
    if(*p || p[1] || p[2]!=0x01) continue;
    u8 sid=p[3];
    if(sid==0xba) {
      i64 scr=unpack(p+4,SCR_MASK);
      pack(p+4,SCR_MASK,scr+delta);
    } else if(sid>=0xc0 && sid<=0xef || sid==0xbd) {
      u8 flags=p[7];
      p+=9;
      if(flags&0x80) {
        i64 pts=unpack(p,PTS_MASK);
        pack(p,PTS_MASK,pts+delta);
        p+=5;
      }
      if(flags&0x40) {
        i64 dts=unpack(p,DTS_MASK);
        pack(p,DTS_MASK,dts+delta);
        p+=5;
      }
      if(flags&0x20) {
        i64 escr=unpack(p,ESCR_MASK);
        pack(p,ESCR_MASK,escr+delta);
        p+=6;
      }
      break;
    }
  }
}

void trim_mpeg2(u8* p) {
  u8 *q=p+pack_size-2;
  for(;q>p;q--) {
    if(*q==0xff && (q[1]&0xfe)==0xfc) {
      *q=0xf0; return;
    };
  }
}
  
u8 *sys_block=0;
u8 *sys_scr=0;
u8 *sys_pad=0;
u8 *end_block=0;

bool is_sid(const u8 *p,u8 sid) {
  return p[0]==0 && p[1]==0 && p[2]==0x01 && p[3]==sid;
}

blockinfo_t parse_block(const u8 *p,const blockinfo_t *seqhd=0) {
  static u64 vhd=~0ull;
  static u32 ahd=0;
  static u64 voffs=0,aoffs=0;
  static vector<u64> sloffs(0x25);

  const u8 *pbeg=p;
  const u8 *pend=p+pack_size;
  blockinfo_t bi;

  if(is_sid(p,0xbe)) p+=6+(p[4]<<8)+p[5]; // Padding stream
  if(p>=pend) return bi;
  if(is_sid(p,0xb9)) { // Program end
    dmp(3,"found end block.\n");
    bi.sid=0xb9; 
    if(!end_block) {
      dmp(3,"found end block.\n");
      end_block=new u8[pack_size];
    } else {
      warn("found additional end block. updating.\n");
    }
    memcpy(end_block,pbeg,pack_size);
    return bi; 
  }
  if(!is_sid(p,0xba)) { // Pack header
    warn("PACK header not found\n");
    return bi;
  }
  p+=4;
  bi.scr=(p[0]&0x38)<<27|(p[0]&0x03)<<28|p[1]<<20|(p[2]&0xf8)<<12|(p[2]&0x03)<<13|p[3]<<5|p[4]>>3;
  bi.mux=(p[6]<<14|p[7]<<6|p[8]>>2)*50;
  p+=10+(p[9]&7);
  if(is_sid(p,0xbb)) { // System Header
    if(!sys_block) {
      dmp(3,"found system header.\n");
      sys_block=new u8[pack_size];
      memcpy(sys_block,pbeg,pack_size);
      u8 *q=sys_block+(p-pbeg)+6+(p[4]<<8)+p[5];
      int npad=pack_size-((p-pbeg)+12+(p[4]<<8)+p[5]);
      *q++=0; *q++=0; *q++=1; *q++=0xbe; *q++=npad>>8; *q++=npad&0xff;
      while(q<sys_block+pack_size) *q++=0xff;
    } else {
      dmp(2,"found additional system header. ignoring.\n");
    }
    p+=6+(p[4]<<8)+p[5];
  }
  if(is_sid(p,0xbe)) p+=6+(p[4]<<8)+p[5]; // Padding stream
  if(p>=pend) return bi;
  if(p[0] || p[1] || p[2]!=0x01 || p[3]<0xbc) { 
    warn("PES prefix not found\n");
    return bi;
  }
  bi.sid=p[3];
  int pes_len=p[4]<<8|p[5];
  p+=6;
  if(p+pes_len>pend) warn("block length mismatch %x+%x!=%x\n",p,pes_len,pend);
  if(bi.sid>=0xbd && bi.sid<=0xef) {
    int pes_hlen=p[2];
    dmp(3,"found PES type %02x, flags=%02x%02x, hlen=%d\n",bi.sid,p[0],p[1],pes_hlen);
    if((p[1]&0x80) && pes_hlen>=5)
      bi.pts=(p[3]&0x0e)<<29|p[4]<<22|(p[5]&0xfe)<<14|p[6]<<7|p[7]>>1;
    p+=3+pes_hlen;
  } else {
    warn("unknown PES type %02x\n",bi.sid);
    return bi;
  }

  if(bi.sid==0xe0) {		// 1st video stream
    u64 seq=0,gop=0,pic=0;
    if(seqhd) {
      seq=seqhd->form;
    } else {
      vhd=~0ull; 
      voffs=0; 
      sloffs.clear(); 
    }
    for(;p<pend;vhd<<=8,vhd|=*p++,voffs++) {
      u32 sc=vhd>>32;
      if(seq && sc>=0x00000101 && sc<=0x000001af) {
        dmp(4,"scanline %d found at offset %d.\n",(sc&0xff),voffs);
        sloffs.push_back(voffs);
        if(sloffs.size()!=(sc&0xff)) {
          warn("unexpected scanline %d.\n",(sc&0xff));
          sloffs.clear();
          break;
        }
      } else if(seq && sloffs.size() && (sc&0xffffff00)==0x00000100) {
        bi.val=lbox_thres;
        int nsl=seq>>12&0xff;
        int sk=nsl/4,ls=lbox_slices;
        if(sk<=ls) error(1,0,"letterbox width (%d) too large for format (%d slices).\n",ls,nsl);
        if(sloffs.size()!=nsl) {
          warn("unexpected number of scanlines (%zd != %d)\n",sloffs.size(),nsl);
          sloffs.clear();
          break;
        }
        sloffs.push_back(voffs);
        int lbox_bytes=sloffs[ls]-sloffs[0]+voffs-sloffs[nsl-ls]-4*ls;
        int frame_bytes=sloffs[nsl-sk]-sloffs[sk]-4*(nsl-2*sk);
        bi.val=(100*(nsl-2*sk)*lbox_bytes)/(2*ls*frame_bytes);
        break;
      } else if(sc==0x000001b3 && !seq) {
        seq=vhd;
        dmp(4,"SEQ header found.\n");
      } else if(sc==0x000001b8 && !gop) {
        gop=vhd; 
        dmp(4,"GOP header found.\n");
      } else if(sc==0x00000100 && !pic) {
        pic=vhd; 
        dmp(4,"PIC header found.\n");
      }
    }
    if(seq && gop && (pic&0x380000)==0x80000 && sloffs.size()) {
      bi.form=seq;
      int tsn=(pic>>22)&0x3ff;
      int fr=seq&0xf;
      bi.pts-=tsn*vframe_dur[fr];
    }
    bi.cnt=sloffs.size();
  } else if(bi.sid==0xc0) {	// 1st audio stream
    if(!seqhd) { ahd=0; aoffs=0; }
    double stereo_factor=0;
    for(;p<pend;ahd<<=8,ahd|=*p++,aoffs++) {
      if((ahd&0xfff80000)==0xfff80000) {
        int l=get_audio_len(ahd);
        if(!l || p+l>pend || p[l-4]!=0xff || p[l-3]!=p[-3] || (p[l-2]^p[-2])&0xfc) continue;
        if(bi.form && (bi.form^ahd)&0xfffffccf)
          warn("audio format mismatch\n");
        double stf=get_stereo_factor(ahd,p);
        if(stf>=0) {
          if(!bi.cnt) {
            bi.form=0xfffffeff&ahd; // clear private bit
          }
          stereo_factor+=stf;
          bi.cnt++;
        }
        p+=l-4;
      }
    }
    if(bi.cnt) bi.val=min<int>(65535,lrint(stereo_factor/bi.cnt));
  }
  return bi;
}

vob_container vobs;
map<int,blockinfo_t> binf;
typedef map<int,blockinfo_t>::iterator binf_iter;

long n_blocks=0;
long n_seq=0;
long n_video=0;
long n_audio=0;

blockinfo_t get_block(int i) {
  blockinfo_t bi;
  if(i<0 || i>=vobs.blocks()) error(1,0,"requested block %d is out of range\n",i);
  if(vobs.visited(i)) {
    binf_iter it=binf.find(i);
    if(it==binf.end()) 
      return bi;
    else
      return it->second;
  } else {
    bi=parse_block(vobs.block(i));
    n_blocks++;
    if(bi.is_video()) n_video++;
    if(bi.is_audio()) n_audio++;
    if(bi.is_seq()) {
      blockinfo_t bi2=bi;
      int j=i;
      n_seq++;
      while((bi2.is_audio() || bi2.cnt && bi2.cnt<=bi.scanlines()) && ++j<vobs.blocks()) {
        bool visited=vobs.visited(j);
        bi2=parse_block(vobs.block(j),&bi);
        if(bi2.is_audio() && !visited) {
          n_audio++;
          binf[j]=bi2;
        }
      }
      if(bi2.is_video() && bi2.cnt==bi.scanlines()+1) {
        bi.cnt=bi2.cnt;
        bi.val=bi2.val;
      } else {
        bi.cnt=0;
        bi.val=lbox_thres;
      }
    }
    if(bi.is_as() || i==0) binf[i]=bi;
    return bi;
  }
}

long n_dist=0;
double sum_dist=0;
double sum2_dist=0;
long n_sync=0; 

const double sds=3; // skip confidence interval in SDs
const double npred=5;
const double decay=1-1/npred;

enum skip_t { SKIP_NEVER, SKIP_NOBOUND, SKIP_ALWAYS, SKIP_AUTO };

// check blocks in [a,b)  
// skip_mode: 0 = complete scan, 1 = not at boundaries, 2 = always skip

long scan_blocks(long a,long b,skip_t skip_mode=SKIP_ALWAYS) { // skip_mode: 0 = complete scan, 1 = not at boundaries, 2 = always skip
  long nb=vobs.blocks();
  if(a<0) a=0;
  if(b>nb) b=nb;
  long ls=a-1,l0=a-1,l1=a-1;
  long ns=0;
  double d1= n_dist>=npred ? sum_dist*npred/n_dist : sum_dist/npred;
  double d2= n_dist>=npred ? sum2_dist*npred/n_dist : sum2_dist/npred;
  double d=d1/npred;
  double sd=sqrt(max<double>(1,d2/npred-d*d));
if(skip_mode==SKIP_AUTO) dmp(2,"scanning %ld..%ld\n",a,b);
  if(skip_mode==SKIP_AUTO) 
    skip_mode= n_dist>npred && b-a>3*d ? SKIP_ALWAYS : SKIP_NEVER;
    
  for(long i=a;i<b;) {
    blockinfo_t bi=get_block(i);
    if(bi.is_as())
      while(i && get_block(i-1).is_as()) i--;
    int s=i;
    while(s<nb && !get_block(s).is_seq()) s++;
    if(s>=nb) break;
    long i0=s,i1=s;
    while(i1<nb-1 && get_block(i1+1).is_as()) i1++;
    while(i0 && get_block(i0-1).is_as()) i0--;
    double dist=s-ls;
    if(i>l1+1) { // have we skipped blocks?
      if(dist>d+sds*sd) { n_sync++; i=l1+1; continue; }
    }
    ns++; ls=s; l0=i0; l1=i1;
    n_dist++;
    sum_dist+=dist;
    sum2_dist+=dist*dist;
    if(skip_mode==SKIP_NEVER) { i=i1+1; continue; }
    d1=decay*d1+dist;
    d2=decay*d2+dist*dist;
    d=d1/npred;
    sd=sqrt(max<double>(1,d2/npred-d*d));
    long i_skip=i0+lrint(d-sds*sd);
    if(skip_mode==SKIP_NOBOUND && (ns<=1 || i_skip>=b && i1+1<b)) { i=i1+1; continue; }
    i=max<long>(i0+lrint(d-sds*sd),i1+1);
  }
  return ns;
}

format_t get_format(binf_iter a,binf_iter b,long context=0) { // [a,b)
  format_t f={0,0,0,0};
  long na=0,nv=0,nlb=0,nsm=0,vlb=0,vsm=0;
  for(binf_iter it=a;it!=b;it++) {
    blockinfo_t bi=it->second;
    if(bi.is_audio()) {
      if(!na || !f.af) f.af=bi.format(); else if(f.af!=bi.format() && bi.format()) f.af=na;
      na++; nsm+=bi.is_stmono(); vsm+=bi.val;
    } else if(bi.is_seq()) {
      if(!nv) f.vf=bi.format(); else if(f.vf!=bi.format()) f.vf=0;
      nv++; nlb+=bi.is_lbox(); vlb+=bi.val;
    }
  }
  f.stmono= (chk_mono && vsm<na*mono_thres);
  f.lbox=   (chk_lbox && vlb<nv*lbox_thres);

  if(context && (chk_lbox || chk_mono)) {
    binf_iter aa=a,bb=b;
    while(aa!=binf.begin() && a->first-aa->first<context && !(aa->second.is_seq() && aa->second.format()!=f.vf)) aa--;
    while(bb!=binf.end() && bb->first-b->first<context && !(bb->second.is_seq() && bb->second.format()!=f.vf)) bb++;
    na=0,nv=0,nlb=0,nsm=0,vlb=0,vsm=0;
    for(binf_iter it=++aa;it!=bb;it++) {
      blockinfo_t bi=it->second;
      if(bi.is_audio()) {
        na++; nsm+=bi.is_stmono(); vsm+=bi.val;
      } else if(bi.is_seq()) {
        nv++; nlb+=bi.is_lbox(); vlb+=bi.val;
      }
    }
    if(chk_mono) {
      if(vsm<na*mono_thres && 2*nsm>na) f.stmono=1;
      if(vsm>na*mono_thres && 2*nsm<na) f.stmono=0;
    }
    if(chk_lbox) {
      if(vlb<nv*lbox_thres && 2*nlb>nv) f.lbox=1;
      if(vlb>nv*lbox_thres && 2*nlb<nv) f.lbox=0;
    }
  }
  return f;
}


void write_file(const char *fname,long a,long b) { // [a,b)
  FILE *f=stdout;
  if(strcmp("-",fname)) {
    f=fopen(fname,"wb");
    if(!f) error(1,errno,"cannot open %s\n",fname);
  }
  if(a && sys_block) {
    memcpy(sys_block,vobs.block(a-1),10); // copy scr
    if(fwrite(sys_block,pack_size,1,f)!=1) error(1,errno,"cannot write sys header\n");
  }
  for(long i=a;i<b;i++) {
    if(fwrite(vobs.block(i),pack_size,1,f)!=1) error(1,errno,"error writing %s\n",fname);
  }
  fclose(f);
}

int main(int argc,char **argv) {
  parse_options(argc,argv);
  if(do_trunc>=0) {
    if(optind!=argc-1) error(1,0,"multiple files given with -T option\n");
    if(ofile) error(1,0,"output file given with -T option\n");
    if(!chunks.empty()) error(1,0,"chunks selected with -T option\n");
    warn("truncate %s to size %lld (%lld blocks)\n",argv[optind],do_trunc*pack_size,do_trunc);
    if(truncate(argv[optind],do_trunc*pack_size)) error(1,errno,"truncation failed.\n");
    return 0;
  }
  if(create_cachedir) {
    if(mkdir(CACHE_DIR,0750) && errno!=EEXIST) error(1,errno,"cannot create directory .vobcut.\n"); 
    if(optind>=argc) exit(0);
  }
  if(optind>=argc) usage();
  {
    struct stat st;
    if(stat(CACHE_DIR,&st) || !S_ISDIR(st.st_mode) && !S_ISLNK(st.st_mode))
      read_cachefile=write_cachefile=0;
  }
  while(optind<argc) vobs.add(argv[optind++]);
  string cachefile=vobs.cachefile();

  long nb=vobs.blocks();
  blockinfo_t firstblock=get_block(0);
  blockinfo_t lastblock=get_block(nb-1);
  double dur=(lastblock.scr-firstblock.scr)/90000.0;
  double blps=nb/dur;
  double bitps=8*pack_size*blps;
  long vcontext=lrint(vote_window*blps/2);
  dmp(1,"%ld blocks, %.2fs, %.1f blocks/s, %.0f kbit/s",nb,dur,blps,bitps/1000);
  if(chk_lbox) dmp(1,", lbox <%d%%",lbox_thres);
  if(chk_mono) { if(sample_mono_thres) dmp(1,", stmono auto"); else dmp(1,", stmono <%d",mono_thres); }
  if(chk_lbox || chk_mono) dmp(1,", context %ld blocks",vcontext);
  dmp(1,"\n");

  long cached_blocks=0;
  if(read_cachefile) { 
    FILE *cache=fopen(cachefile.c_str(),"rb");
    long cpar[8];
    if(cache && fread(cpar,sizeof(long),8,cache)==8 && cpar[0]==CACHE_MAGIC && cpar[1]==sizeof(blockinfo_t) && cpar[2]==pack_size && cpar[3]==nb) {
      dmp(1,"reading cache file %s\n",cachefile.c_str());
      long nc_tot=0,nc_bi=0;
      unsigned u;
      blockinfo_t bi;
      while(fread(&u,sizeof(unsigned),1,cache)) {
        if(u&0x80000000) {
          if(!fread(&bi,sizeof(blockinfo_t),1,cache)) break;
          nc_bi++;
          u&=0x7fffffff;
          binf[u]=bi;
        }
        nc_tot++;
        vobs.visit_block(u);
      }
      cached_blocks=vobs.n_loaded;
      n_blocks=cpar[4];
      n_seq=cpar[5];
      n_video=cpar[6];
      n_audio=cpar[7];
      dmp(1,"%ld blocks with %ld blockinfos read.\n",nc_tot,nc_bi);
    } else {
      if(cache) dmp(0,"ignoring invalid cache file %s\n",cachefile.c_str());
    }
    if(cache) fclose(cache);
  }

  if(dump) {
    int gap=0;
    int naudio=0;
    int nvideo=0;
    for(int i=0;i<vobs.blocks();i++) {
      blockinfo_t bi=get_block(i);
      if(bi.form) {
        if(gap) dmp(2,"\n");
        gap=0;
        const char *typstr=bi.sid==0xe0 ? "V" : bi.sid==0xc0 ? "A" : " ";
        const char *hdstr=bi.sid==0xe0 ? vhd_str(bi.form) : bi.sid==0xc0 ? ahd_str(bi.form) : "";
        const char *lboxstr= bi.is_lbox() ? " letterbox":"";
        const char *stmonostr= bi.is_stmono() ? " stmono":"";
        printf("%7d: %s %02x  SCR%9.3f  PTS%9.3f  MUX%6.2f  %2d %5d %s%s%s\n",
            i,typstr,bi.sid,bi.scr/90000.0,bi.pts/90000.0,bi.mux/131072.0,bi.cnt,bi.val,hdstr,lboxstr,stmonostr);
        if(bi.sid==0xe0) nvideo++;
        if(bi.sid==0xc0) {
          naudio+=bi.cnt;
        }
      } else {
        dmp(2,".");
        gap=1;
      }
    }
    printf("%d video sequences, %d audio frames\n",nvideo,naudio);
    return 0;
  }


  long si,sl;
  if(iscan_len==0 || (!iscans && iscan_len*min_iscans*2>dur)) {
    dmp(1,"performing complete scan over %ld blocks (full coverage)...\n",nb);
    iscans=1;
    si=0; sl=nb;
  } else {
    if(!iscans) {
      iscans=lrint(sqrt(2*dur/iscan_len));
      if(iscans<min_iscans) iscans=min_iscans;
      if(iscan_ival && (dur-iscan_len)/iscans>iscan_ival) iscans=lrint((dur-iscan_len)/iscan_ival+1.5);
    }
    if(iscans>1) {
      si=(nb-lrint(iscan_len*blps))/(iscans-1);
      sl=nb-(iscans-1)*si;
    } else {
      si=nb;
      sl=lrint(iscan_len*blps);
    }
    dmp(1,"performing %ld scans over %ld blocks each at %ld-block intervals (%.2f%% coverage) ...\n",iscans,sl,si,100.0*sl*iscans/(double)nb);
  }
  for(long i=0;i<iscans;i++) {
    dmp(2,"scanning blocks %ld..%ld: ",i*si,i*si+sl-1);
    long ns=scan_blocks(i*si,i*si+sl,i==0 || i==iscans-1 ? SKIP_NOBOUND : SKIP_ALWAYS);
    dmp(2,"%ld sequences found.\n",ns);
  }
  
  map<u32,long> aforms;
  map<u32,long> vforms;
  long n_lbox=0;
  long n_stmono=0;
  long a_votes=0,v_votes=0;
  binf_iter vote_beg= (iscans<3) ? binf.begin() : binf.lower_bound(si/2);
  binf_iter vote_end= (iscans<3) ? binf.end() : binf.lower_bound(nb-si/2);
  for(binf_iter it=vote_beg;it!=vote_end;it++) {
    blockinfo_t bi=it->second;
    if(bi.is_audio()) {
      a_votes++;
      aforms[bi.format()]++;
      if(bi.is_stmono()) n_stmono++;
    } else if(bi.is_seq()) {
      v_votes++;
      vforms[bi.format()]++;
      if(bi.is_lbox()) n_lbox++;
    }
  }
  map<u32,long>::iterator amax,vmax;
  if(a_votes) {
    for(map<u32,long>::iterator it=amax=aforms.begin();it!=aforms.end();it++) 
      if(amax->second<it->second) amax=it;
    dmp(1,"audio format: %ld/%ld %s, %ld/%ld st_mono\n",amax->second,a_votes,ahd_str(amax->first),n_stmono,a_votes);
  } else {
    warn("no audio packets found\n");
  }
  if(v_votes) {
    for(map<u32,long>::iterator it=vmax=vforms.begin();it!=vforms.end();it++) 
      if(vmax->second<it->second) vmax=it;
    dmp(1,"video format: %ld/%ld %s, %ld/%ld letterbox\n",vmax->second,v_votes,vhd_str(vmax->first),n_lbox,v_votes);
  } else {
    error(1,0,"no video packets found\n");
  }
  format_t fmt;
  fmt.af=amax->second;
  fmt.vf=vmax->second;
  fmt.lbox=(2*n_lbox>v_votes);
  fmt.stmono=(2*n_stmono>a_votes);
  if(chk_lbox==-1) { chk_lbox=fmt.lbox;   dmp(2,"letterbox detection %sabled.\n",fmt.lbox ? "en" : "dis"); }
  if(chk_mono==-1) { chk_mono=fmt.stmono; dmp(2,"stereo encoded mono detection %sabled.\n",fmt.stmono ? "en" : "dis"); }
  
  if(chk_mono && sample_mono_thres) { 
    if(dur>600 && 10*n_stmono>a_votes && 10*n_stmono<9*a_votes) {
      const long chunk=lrint(min<double>(120,dur/10)*blps);
      multiset<double> stfs;
      for(long pos=0;;pos+=chunk) {
        binf_iter a=binf.lower_bound(pos);
        binf_iter b=binf.lower_bound(pos+chunk);
        if(a==binf.end()) break;
        if(b->first-a->first<chunk/2) continue;
        long n=0;
        double stf=0;
        for(binf_iter it=a;it!=b;it++) {
          blockinfo_t bi=it->second;
          if(!bi.is_audio()) continue;
          n++; stf+=bi.val;
        }
        if(n<10) continue;
        stfs.insert(stf/n);
      }
      double stf_low_limit=mono_thres/2,stf_high_limit=mono_thres*2;
      if(stfs.size()<3 || *(stfs.begin())>stf_low_limit || *(stfs.rbegin())<stf_high_limit)
        warn("inconclusive mono-threshold - auto detection disabled, threshold %g\n",mono_thres);
      else {
        stf_low_limit=*(stfs.begin());
        stf_high_limit=*(stfs.rbegin());
        if(stf_low_limit<0.1*mono_thres) stf_low_limit=0.1*mono_thres;
        if(stf_high_limit>10*mono_thres) stf_high_limit=10*mono_thres;
        mono_thres=cbrt(mono_thres*stf_high_limit*stf_low_limit);
        dmp(1,"sampled stereo factors %g .. %g, %zd chunks, threshold %g\n",
            *(stfs.begin()),*(stfs.rbegin()),stfs.size(),mono_thres);
      }
    } else {
      dmp(1,"insufficient sample data, using default threshold %g\n",mono_thres); 
    }
  }

  dmp(1,"%ld sequences, %ld / %ld / %ld blocks used / loaded / total (%.2f%%)\n\n",n_seq,binf.size(),vobs.n_loaded,vobs.blocks(),100*vobs.covered());

  {
    binf_iter la,lv,it;
    for(it=binf.begin(),la=lv=binf.end();it!=binf.end();it++) {
      blockinfo_t bi=it->second;
      binf_iter &l= bi.is_audio() ? la : lv;
      long nb=binf.size();
      if(l!=binf.end()) {
        blockinfo_t lbi=l->second;
        if(bi.format()!=lbi.format()) {
          dmp(2,"fm "); scan_blocks(l->first+1,it->first,SKIP_AUTO);
          if(nb!=binf.size()) {
            it=l;
            continue;
          }
        } else if(bi.is_special() ^ lbi.is_special()) {
          dmp(2,"ss "); scan_blocks(l->first+1-vcontext,it->first+vcontext,SKIP_AUTO);
          if(nb!=binf.size()) {
            if(vcontext) {
              it=binf.lower_bound(l->first-vcontext);
              if(it==binf.end()) it=binf.begin();
              la=lv=binf.end();
            } else {
              it=l;
            }
            continue;
          }
        }
      }
      l=it;
    }
  }

  map<long,format_t> seq;
  map<u32,long> pts;
  map<long,format_t> part;
  map<long,long> partlen;
  multimap<long,long> chunkmap;
  set<long> is_chunk;
  vector<long> chunk;
  
  {
    if(n_seq<2) error(1,0,"file contains less than two sequences.\n");
    binf_iter a,b,c;
    a=b=c=binf.begin();
    format_t of={0,0,0,0};
    while(c!=binf.end() && !c->second.is_seq()) c++;
    
    for(;a!=binf.end();a=b) {
      if(c!=binf.end()) c++;
      while(c!=binf.end() && !c->second.is_seq()) c++;
      b=c;
      if(c!=binf.end()) {
        u32 pts=c->second.pts;
        binf_iter it=b;
        for(it--;it->second.is_audio() && it->second.pts>=pts;it--) b--;
      }
      format_t f=get_format(a,b,vcontext);
      if(!f.af && b!=binf.end()) { b=a; continue; }
      seq[a->first]=f;
      pts[a->second.pts]=a->first;
      if(f!=of) {
        part[a->first]=of=f;
      }        
    }
    if(remove_limit && part.size()>2) {
      long rlimit=lrint(remove_limit*blps);
      long npart;
      do {
        npart=part.size();
        map<long,format_t>::iterator d=part.begin(),a=d++,b=d++,c=d++;
        for(;a!=part.end() && b!=part.end() && c!=part.end();a++,b++,c++,d++) {
          long al=b->first-a->first;
          long bl=c->first-b->first;
          long cl= d!=part.end() ? d->first-c->first : nb-c->first;
          format_t af=a->second;
          format_t bf=b->second;
          format_t cf=c->second;
          if(a==part.begin() && sim(af,bf) && al<rlimit && al<bl) {
            a->second=bf;
            part.erase(b);
            b=c++; d++;
          }
          if(d==part.end() && (sim(bf,cf) || bf.vf==cf.vf && !cf.af) && cl<rlimit && cl<bl) {
            part.erase(c);
            break;
          }
          if(bl>rlimit || bl>=al && bl>=cl) continue;
          if(!sim(af,bf) && !sim(bf,cf)) continue;
          if(!sim(af,bf) && bl>=cl || !sim(bf,cf) && bl>=al) continue;
          if(af==cf && bl<al && bl<cl) {
            part.erase(b);
            part.erase(c);
            b=d++; c=d++;
          } else if(!sim(af,bf) && sim(bf,cf) && bl<cl) {
            b->second=cf;
            part.erase(c);
            c=d++;
          } else if(sim(af,bf) && !sim(bf,cf) && bl<al) {
            part.erase(b);
            b=c++; d++;
          }
        }
      } while(npart!=part.size() && part.size()>2);
    }
    
    while(!tcuts.empty()) {
      u64 t=llrint(90000ull*tcuts.back());
      tcuts.pop_back();
      map<u32,long>::iterator it1=pts.lower_bound(t),it0=it1;
      if(it1==pts.end()) continue;
      if(it0!=pts.begin()) it0--;
      if(t-it0->first > it1->first-t) it0=it1;
      cuts.push_back(it0->second);
    }
    
    while(!cuts.empty()) {
      long c=cuts.back();
      map<long,format_t>::iterator it=seq.lower_bound(c);
      if(it!=seq.end()) {
        part[it->first]=it->second;
      }
      cuts.pop_back();
    }
    
    long lend=vobs.blocks();
    for(map<long,format_t>::reverse_iterator it=part.rbegin();it!=part.rend();it++) {
      long l=lend-it->first;
      chunkmap.insert(pair<long,long>(l,it->first));
      partlen[it->first]=l;
      lend=it->first;
    }
    while(chunkmap.size()>10) chunkmap.erase(chunkmap.begin());
    for(multimap<long,long>::iterator it=chunkmap.begin();it!=chunkmap.end();it++)
      is_chunk.insert(it->second);
    for(map<long,format_t>::iterator it=part.begin();it!=part.end();it++) {
      if(is_chunk.find(it->first)!=is_chunk.end()) {
        int i=chunk.size();
        chunk.push_back(it->first);
        if(chunks.empty() && it->first==chunkmap.rbegin()->second) chunks.insert(i);
        dmp(1,"%c %d: ",chunks.find(i)!=chunks.end() ? '*' : ' ',i);
      } else {
        dmp(1,"     ");
      }
      dmp(1,"%7ld %5.0fs  %s\n",it->first,partlen[it->first]/blps,it->second.str());
    }
  }
  dmp(1,"     %7ld blocks total.\n",vobs.blocks());

  if(plot) {
    printf("## set plotter option -h 10\n");
    printf("## set plotter option -w 32\n");
    printf("## set plotter option -T %s\n",vobs.filename().c_str());
    printf("## set plotter option -L lbox\n");
    printf("## set plotter option -L stmono\n");
    printf("## set plotter option -l\n");
    printf("## set plotter option -g\n");
    printf("## set plotter option -g\n");
    printf("## set plotter option -g\n");
    for(map<long,format_t>::iterator it=part.begin();it!=part.end();it++) {
      printf("## set plotter option -V %.2f\n",it->first/blps);
    }
    printf("## set plotter option -c\n");
    for(binf_iter it=binf.begin();it!=binf.end();it++) {
      if(!it->second.is_audio()) continue;
      printf("%g %d\n",it->first/blps,it->second.val);
    }
    putchar('\n');
    for(binf_iter it=binf.begin();it!=binf.end();it++) {
      if(!it->second.is_seq()) continue;
      printf("%g %d\n",it->first/blps,it->second.val);
    }
  }
  
  if(chunks.empty() || *(chunks.rbegin())>=chunkmap.size())
    error(1,0,"invalid chunk selection\n");

  part[vobs.blocks()]=part.rbegin()->second;
  seq[vobs.blocks()]=seq.rbegin()->second;
  FILE *f=0;
  if(ofile) {
    if(strcmp("-",ofile)) {
      f=fopen(ofile,"wb");
      if(!f) error(1,errno,"cannot open %s\n",ofile);
    } else {
      f=stdout;
    }
  }
  i64 next_scr=1; 

  for(set<int>::iterator ck=chunks.begin();ck!=chunks.end();ck++) {
    int c=*ck;
    long first_block=chunk[c];
    long last_block=part.upper_bound(first_block)->first-1;
    format_t fmt=part[first_block];
    if(adj_a && c==*(chunks.begin())) {
      long d=abs(lrint(adj_a*blps));
      map<long,format_t>::iterator a=seq.find(first_block),aa=a;
      while(aa!=seq.end() && abs(aa->first-first_block)<=d) {
        if(aa->second.af!=fmt.af) warn("adjust start: audio format mismatch\n");
        if(aa->second.vf!=fmt.vf) warn("adjust start: video format mismatch\n");
        a=aa;
        if(adj_a<0) aa--; else aa++;
      }
      first_block=a->first;
    }
    if(adj_b && c==*(chunks.rbegin())) {
      long d=abs(lrint(adj_b*blps));
      map<long,format_t>::iterator b=seq.find(last_block+1),bb=b;
      while(bb!=seq.end() && abs(bb->first-last_block-1)<=d) {
        if(bb->second.af!=fmt.af) warn("adjust end: audio format mismatch\n");
        if(bb->second.vf!=fmt.vf) warn("adjust end: video format mismatch\n");
        b=bb;
        if(adj_b<0) bb--; else bb++;
      }
      last_block=b->first-1;
    }
    i64 shift=next_scr-binf[first_block].scr;
    if(f && c==*(chunks.begin()) && first_block && sys_block) {
      shift=next_scr-binf[first_block-1].scr;
      memcpy(sys_block,vobs.block(first_block-1),10); // copy scr
      if(calc_timestamps) shift_block(sys_block,shift);
      if(fwrite(sys_block,pack_size,1,f)!=1) error(1,errno,"cannot write sys header\n");
    }
    next_scr=binf[last_block+1].scr+shift;
    if(preview) {
      dmp(1,"preview %schunk %d: blocks %ld .. %ld (%.2f MB)\n",view_bound?"bounderies of ":"",
          c,first_block,last_block,(last_block-first_block+1)*0.00000095367431640625*pack_size);
      long n=lrint(blps*preview_len);
      long a,b,c,d;
      if(view_bound) {
        a=max<long>(0,first_block-n); b=first_block;
        c=last_block; d=min<long>(vobs.blocks(),last_block+n);
      } else {
        a=first_block; b=first_block+n;
        c=last_block-n; d=last_block;
      }
      for(long i=a;i<b;i++) {
        fwrite(vobs.block(i),pack_size,1,stdout);
      }
      for(long i=c;i<=d;i++) {
        fwrite(vobs.block(i),pack_size,1,stdout);
      }
    }
    long trim_block=-1;
    if(trim_audio) {
      set<int>::iterator nck=ck;
      if(++nck!=chunks.end() && chunk[*nck]!=last_block+1) {
        trim_block=last_block;
        while(trim_block>first_block && get_block(trim_block).sid!=0xc0)
          trim_block--;
      }
    }
    if(f) {
      dmp(1,"writing chunk %d: blocks %ld .. %ld (%.2f MB)\n",c,first_block,last_block,(last_block-first_block+1)*0.00000095367431640625*pack_size);
      for(long i=first_block;i<=last_block;i++) {
        u8 *p=vobs.block(i);
        if(i==trim_block) trim_mpeg2(p);
        if(calc_timestamps) shift_block(p,shift);
        if(fwrite(p,pack_size,1,f)!=1) error(1,errno,"error writing %s\n",ofile);
      }
    }
  }

  if(write_cachefile && vobs.n_loaded>cached_blocks) {
    dmp(1,"writing cache file %s\n",cachefile.c_str());
    FILE *cache=fopen(cachefile.c_str(),"wb");
    long cpar[8];
    cpar[0]=CACHE_MAGIC;
    cpar[1]=sizeof(blockinfo_t);
    cpar[2]=pack_size;
    cpar[3]=nb;
    cpar[4]=n_blocks;
    cpar[5]=n_seq;
    cpar[6]=n_video;
    cpar[7]=n_audio;
    long nc_tot=0,nc_bi=0;
    if(fwrite(cpar,sizeof(long),8,cache)!=8) error(1,errno,"error writing cache\n");
    for(long i=0;i<nb && i<=0x7fffffff;i++) {
      unsigned u=i;
      if(vobs.visited(i)) {
        nc_tot++;
        binf_iter it=binf.find(i);
        if(it!=binf.end()) {
          u|=0x80000000;
          fwrite(&u,sizeof(unsigned),1,cache);
          fwrite(&(it->second),sizeof(blockinfo_t),1,cache);
          nc_bi++;
        } else {
          fwrite(&u,sizeof(unsigned),1,cache);
        }
      }
    }
    fclose(cache);
    dmp(1,"%ld blocks with %ld blockinfos written.\n",nc_tot,nc_bi);
  }    

  return 0;
}
